@@ -13,6 +13,7 @@ web service.
 from __future__ import absolute_import, print_function, unicode_literals
 from multiprocessing import cpu_count
 from multiprocessing.dummy import Pool as ThreadPool
+import asyncio
 import datetime
 import json
 import os
@@ -38,7 +39,7 @@ from elliottlib.util import validate_email_address, red_print, major_from_branch
 from elliottlib.util import green_print, red_prefix, minor_from_branch
 from elliottlib.util import yellow_print, yellow_prefix, exit_unauthorized, release_from_branch
 from elliottlib.util import progress_func, pbar_header
-from elliottlib.cli.common import cli, use_default_advisory_option, find_default_advisory
+from elliottlib.cli.common import cli, use_default_advisory_option, find_default_advisory, click_coroutine
 
 # cli commands
 from elliottlib.cli.list_cli import list_cli
@@ -84,7 +85,7 @@ pass_runtime = click.make_pass_decorator(Runtime)
 @use_default_advisory_option
 @click.option("--mode",
               required=True,
-              type=click.Choice(['list', 'sweep', 'diff']),
+              type=click.Choice(['list', 'sweep', 'diff', 'qe']),
               default='list',
               help='Mode to use to find bugs')
 @click.option("--status", 'status',
@@ -110,8 +111,12 @@ pass_runtime = click.make_pass_decorator(Runtime)
 @click.option("--into-default-advisories",
               is_flag=True,
               help='attaches bugs found to their correct default advisories, e.g. operator-related bugs go to "extras" instead of the default "image", bugs filtered into "none" are not attached at all.')
+@click.option("--noop", "--dry-run",
+              is_flag=True,
+              default=False,
+              help="Don't change anything")
 @pass_runtime
-def find_bugs(runtime, advisory, default_advisory_type, mode, status, id, from_diff, flag, report, into_default_advisories):
+def find_bugs(runtime, advisory, default_advisory_type, mode, status, id, from_diff, flag, report, into_default_advisories, noop):
     """Find Red Hat Bugzilla bugs or add them to ADVISORY. Bugs can be
 "swept" into the advisory either automatically (--mode sweep), or by
 manually specifying one or more bugs using --mode list and the --id option.
@@ -130,6 +135,10 @@ mode you must provide a list of IDs to attach with the --id option.
 DIFF: For this use case, you must provide the --between option using two
 URLs to payloads.
 
+QE: Find MODIFIED bugs for the target-releases, and set them to ON_QA.
+The --group option MUST be provided. Cannot be used in combination
+with --into-default-advisories, --add, --into-default-advisories
+
 Using --use-default-advisory without a value set for the matching key
 in the build-data will cause an error and elliott will exit in a
 non-zero state. Use of this option silently overrides providing an
@@ -162,6 +171,11 @@ advisory with the --add option.
 
 \b
     $ elliott --group=openshift-4.1 --mode sweep --use-default-advisory rpm
+
+    Find bugs for 4.6 that are in MODIFIED state, and set them to ON_QA:
+
+\b
+    $ elliott --group=openshift-4.6 --mode qe
 """
     if mode != 'list' and len(id) > 0:
         raise click.BadParameter("Combining the automatic and manual bug attachment options is not supported")
@@ -175,6 +189,9 @@ advisory with the --add option.
     if sum(map(bool, [advisory, default_advisory_type, into_default_advisories])) > 1:
         raise click.BadParameter("Use only one of --use-default-advisory, --add, or --into-default-advisories")
 
+    if mode == 'qe' and sum(map(bool, [advisory, default_advisory_type, into_default_advisories])) > 0:
+        raise click.BadParameter("--mode=qe does not operate on an advisory. Do not specify any of `--use-default-advisory`, `--add`, or `--into-default-advisories`")
+
     runtime.initialize()
     bz_data = runtime.gitdata.load_data(key='bugzilla').data
     bzapi = elliottlib.bzutil.get_bzapi(bz_data)
@@ -182,8 +199,10 @@ advisory with the --add option.
     if default_advisory_type is not None:
         advisory = find_default_advisory(runtime, default_advisory_type)
 
-    if mode == 'sweep':
-        green_prefix("Searching for bugs with target release(s):")
+    if mode == 'sweep' or mode == 'qe':
+        if mode == 'qe':
+            status = ['MODIFIED']
+        green_prefix(f"Searching for bugs with status {' '.join(status)} and target release(s):")
         click.echo(" {tr}".format(tr=", ".join(bz_data['target_release'])))
         bugs = elliottlib.bzutil.search_for_bugs(bz_data, status, verbose=runtime.debug)
     elif mode == 'list':
@@ -229,6 +248,18 @@ advisory with the --add option.
         green_prefix("Found {} bugs:".format(len(bugs)))
     click.echo(" {}".format(", ".join([str(b.bug_id) for b in bugs])))
 
+    if mode == 'qe':
+        for bug in bugs:
+            elliottlib.bzutil.set_state(bug, 'ON_QA', noop=noop)
+
+    if len(flag) > 0:
+        for bug in bugs:
+            for f in flag:
+                if noop:
+                    click.echo(f'Would have updated bug {bug.id} by setting flag {f}')
+                    continue
+                bug.updateflags({f: "+"})
+
     if report:
         green_print("{:<8s} {:<25s} {:<12s} {:<7s} {:<10s} {:60s}".format("ID", "COMPONENT", "STATUS", "SCORE", "AGE", "SUMMARY"))
         for bug in bugs:
@@ -241,13 +272,8 @@ advisory with the --add option.
                                                                                     days_ago,
                                                                                     bug.summary[:60]))
 
-    if len(flag) > 0:
-        for bug in bugs:
-            for f in flag:
-                bug.updateflags({f: "+"})
-
     if advisory and not default_advisory_type:  # `--add ADVISORY_NUMBER` should respect the user's wish and attach all available bugs to whatever advisory is specified.
-        elliottlib.errata.add_bugs_with_retry(advisory, bugs)
+        elliottlib.errata.add_bugs_with_retry(advisory, bugs, noop=noop)
         return
 
     # If --use-default-advisory or --into-default-advisories is given, we need to determine which bugs should be swept into which advisory.
@@ -266,11 +292,11 @@ advisory with the --add option.
         impetus_bugs["image"] = {b for b in bugs if b.component not in extra_components}
 
     if default_advisory_type and impetus_bugs.get(default_advisory_type):
-        elliottlib.errata.add_bugs_with_retry(advisory, impetus_bugs[default_advisory_type])
+        elliottlib.errata.add_bugs_with_retry(advisory, impetus_bugs[default_advisory_type], noop=noop)
     elif into_default_advisories:
         for impetus, bugs in impetus_bugs.items():
             if bugs:
-                elliottlib.errata.add_bugs_with_retry(runtime.group_config.advisories[impetus], bugs)
+                elliottlib.errata.add_bugs_with_retry(runtime.group_config.advisories[impetus], bugs, noop=noop)
 
 
 #
@@ -597,7 +623,8 @@ def find_cve_trackers(runtime, cve, status):
 @click.argument("payload")
 @click.argument('advisory', type=int)
 @click.pass_context
-def verify_payload(ctx, payload, advisory):
+@click_coroutine
+async def verify_payload(ctx, payload, advisory):
     """Cross-check that the builds present in PAYLOAD match the builds
 attached to ADVISORY. The payload is treated as the source of
 truth. If something is absent or different in the advisory it is
@@ -669,18 +696,19 @@ written out to summary_results.json.
 
     green_prefix("Looping over payload images: ")
     click.echo("{} images to check".format(len(payload_json['references']['spec']['tags'])))
-    for image in payload_json['references']['spec']['tags']:
+    cmds = [['oc', 'image', 'info', '-o', 'json', tag['from']['name']] for tag in payload_json['references']['spec']['tags']]
+
+    green_prefix("Querying image infos...")
+    cmd_results = await asyncio.gather(*[exectools.cmd_gather_async(cmd) for cmd in cmds])
+
+    for image, cmd, cmd_result in zip(payload_json['references']['spec']['tags'], cmds, cmd_results):
         click.echo("----")
-        green_prefix("Getting payload image metadata: ")
-        click.echo("{}".format(image['from']['name']))
-        pullspec = image['from']['name']
         image_name = image['name']
-        pullspec_cmd = 'oc image info {} -o json'.format(pullspec)
-        rc, stdout, stderr = exectools.cmd_gather(pullspec_cmd)
+        rc, stdout, stderr = cmd_result
         if rc != 0:
             # Probably no point in continuing.. can't contact brew?
             red_prefix("Unable to run oc image info: ")
-            click.echo("out={}  ; err={}".format(stdout, stderr))
+            red_print(f"cmd={cmd!r}, out={stdout}  ; err={stderr}")
             exit(1)
 
         image_info = json.loads(stdout)
@@ -707,10 +735,13 @@ written out to summary_results.json.
 
     missing_in_errata = {}
     payload_doesnt_match_errata = {}
+    in_other_advisories = {}
     output = {
         'missing_in_advisory': missing_in_errata,
         'payload_advisory_mismatch': payload_doesnt_match_errata,
+        "in_other_advisories": in_other_advisories,
     }
+
     green_prefix("Analyzing data: ")
     click.echo("{} images to consider from payload".format(len(all_payload_nvrs)))
 
@@ -728,6 +759,20 @@ written out to summary_results.json.
                 'errata': all_advisory_nvrs[image]
             }
 
+    if missing_in_errata:
+        green_print(f"Checking if {len(missing_in_errata)} missing images are shipped...")
+        nvrs = list(missing_in_errata.values())
+        tag_lists = elliottlib.brew.get_builds_tags(nvrs)
+        for nvr, tags in zip(nvrs, tag_lists):
+            name = nvr.rsplit("-", 2)[0]
+            if any(map(lambda tag: tag["name"].endswith('-released'), tags)):
+                green_print(f"Build {nvr} is shipped. Skipping...")
+                del missing_in_errata[name]
+            elif any(map(lambda tag: tag["name"].endswith('-pending'), tags)):
+                green_print(f"Build {nvr} is in another advisory.")
+                del missing_in_errata[name]
+                in_other_advisories[name] = nvr
+
     green_print("Summary results:")
     click.echo(json.dumps(output, indent=4))
     with open('summary_results.json', 'w') as fp:
