@@ -388,7 +388,7 @@ def parse_exception_error_message(e):
     return [int(b.split('#')[1]) for b in re.findall(r'Bug #[0-9]*', str(e))]
 
 
-def add_bugs_with_retry(advisory, bugs, retried=False):
+def add_bugs_with_retry(advisory, bugs, retried=False, noop=False):
     """
     adding specified bugs into advisory, retry 2 times: first time
     parse the exception message to get failed bug id list, remove from original
@@ -412,16 +412,6 @@ def add_bugs_with_retry(advisory, bugs, retried=False):
         retry_times=1 if retried is False else 2
     ))
 
-    # Due to a limitation of Errata Tool, a bug cannot be attached to an advisory
-    # if it has state ON_QA. This might change, see https://projects.engineering.redhat.com/browse/CLOUDWF-3104
-    for bug in bugs:
-        if bug.status == 'ON_QA':
-            try:
-                bzutil.set_state(bug, 'MODIFIED')
-            except Exception as e:
-                # Accept exception, it will fail while adding the bugs, after which it will be retried
-                logger.warning(f'Could not change BZ {bug.id} to ON_QA: {e.message}, {e.args}')
-
     print(" {advs}".format(advs=advs))
     try:
         advs.addBugs([bug.id for bug in bugs])
@@ -432,7 +422,7 @@ def add_bugs_with_retry(advisory, bugs, retried=False):
             block_list = parse_exception_error_message(e)
             retry_list = [x for x in bugs if x not in block_list]
             if len(retry_list) > 0:
-                add_bugs_with_retry(advisory, retry_list, retried=True)
+                add_bugs_with_retry(advisory, retry_list, retried=True, noop=noop)
         else:
             raise exceptions.ElliottFatalError(getattr(e, 'message', repr(e)))
 
